@@ -13,8 +13,8 @@ from svolfit.models.Heston_utils import Heston_lncondassetprob,Heston_griddefs,H
 #---------------------------------------------
 
 class Heston_tree(svol_model):
-    def __init__( self, series,dt, model, method ):
-        super().__init__(series,dt, model, method)
+    def __init__( self, series,dt, model, method,options ):
+        super().__init__(series,dt, model, method,options)
 
         mu=0.0
         sigma=0.1
@@ -34,30 +34,117 @@ class Heston_tree(svol_model):
 #                 [hmu, hsigma, rho, alpha, xi,u0]
         self.workingpars_bounds=[(-0.5,0.5), (0.05, 0.5), (-0.9,0.9), (alpha_min, 20.0), (0.1, 3.5),(0.1,3.0)]
 
+        if 'init' in options:
+            self.initpars_reporting(options['init'])
+
 # precalculate anything that can absolutely be reused:
 #TODO: ugly!!
         Nret=self.Nobs-1
-        self.yasset=np.log( self.series[1:Nret+1]/self.series[0:Nret] )
-        self.upath=np.zeros(self.Nobs)
+        if(Nret>0):
+            self.yasset=np.log( self.series[1:Nret+1]/self.series[0:Nret] )
+            self.upath=np.zeros(self.Nobs)
 
 #TODO: expose these as options at some point?
-        self.ProbFactor =1.09
-        self.NormProbCalc='TDist'
-        
-        if( self.NormProbCalc == 'Normal' ):
-            sigma_base=np.std(self.yasset,ddof=1)
-            mu_base=np.mean(self.yasset)
-            self.cprob_base=norm.pdf( self.yasset, loc=mu_base, scale=sigma_base )
-        elif( self.NormProbCalc == 'TDist' ):
-            (t_df,t_loc,t_scale)=tdist.fit(self.yasset)
-            self.cprob_base=tdist.pdf(self.yasset,df=t_df,loc=t_loc,scale=t_scale)
-        else:
-            print('Unknown NormProbCalc -- Handle')
-    # allow for a multiplier:
-        self.cprob_base*=self.ProbFactor
+            self.ProbFactor =1.09
+            self.NormProbCalc='TDist'
+            
+            if( self.NormProbCalc == 'Normal' ):
+                sigma_base=np.std(self.yasset,ddof=1)
+                mu_base=np.mean(self.yasset)
+                self.cprob_base=norm.pdf( self.yasset, loc=mu_base, scale=sigma_base )
+            elif( self.NormProbCalc == 'TDist' ):
+                (t_df,t_loc,t_scale)=tdist.fit(self.yasset)
+                self.cprob_base=tdist.pdf(self.yasset,df=t_df,loc=t_loc,scale=t_scale)
+            else:
+                print('Unknown NormProbCalc -- Handle')
+        # allow for a multiplier:
+            self.cprob_base*=self.ProbFactor
 
         return
 
+    def initpars_reporting(self,pardict):
+
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+        u0=self.workingpars[5]
+
+        theta=sigma*sigma
+        eta=xi*sigma
+        v0=theta*u0
+        
+        for x in pardict:
+            if( x=='mu' ):
+                mu=pardict[x]
+            if( x=='theta' ):
+                theta=pardict[x]
+            if( x=='rho' ):
+                rho=pardict[x]
+            if( x=='alpha' ):
+                alpha=pardict[x]
+            if( x=='eta' ):
+                eta=pardict[x]
+            if( x=='v0' ):
+                v0=pardict[x]
+
+        sigma=np.sqrt(theta)
+        xi=eta/sigma
+        u0=v0/theta
+        
+        self.workingpars[0]=mu
+        self.workingpars[1]=sigma
+        self.workingpars[2]=rho
+        self.workingpars[3]=alpha
+        self.workingpars[4]=xi
+        self.workingpars[5]=u0
+
+        return
+
+    def get_structure(self):
+        assetname='asset'
+        variancename='variance'
+
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        u0=self.workingpars[5]
+
+        corrmatrix=np.array([[1.0,rho],[rho,1.0]])
+#TODO: best choice based on pars?
+        Nperstep=4
+
+        assetval=1.0
+        varianceval=u0*sigma*sigma
+
+        return assetname,assetval,variancename,varianceval,corrmatrix,Nperstep
+
+    def sim_step(self,asset,variance,Zs):
+
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+
+        theta=sigma*sigma
+        eta=xi*sigma
+        
+        Nperstep=np.shape(Zs)[1]
+        sim_asset=np.log(asset)
+        sim_variance =variance       
+        
+        dt=self.dt/Nperstep
+        vmin=1.0e-12
+        
+        for cc in range(0,Nperstep):
+            sim_asset+=(mu-0.5*sim_variance)*dt+np.sqrt(sim_variance*dt)*Zs[0,cc,:]
+            sim_variance=np.power(np.sqrt(sim_variance)+0.5*eta*np.sqrt(dt)*Zs[1,cc,:],2)+alpha*(theta-sim_variance)*dt-eta*eta*dt/4.0
+            sim_variance=np.maximum(sim_variance,vmin)
+
+        sim_asset=np.exp(sim_asset)
+        return sim_asset,sim_variance
+
     def get_reportingpars(self):
         super().get_reportingpars()
 
@@ -84,18 +171,18 @@ class Heston_tree(svol_model):
     
         vpath=sigma*sigma*self.upath
 
-        ret['mu']=mu
-        ret['theta']=theta
-        ret['rho']=rho
-        ret['alpha']=alpha
-        ret['eta']=eta
-        ret['q']=q
+        ret['rep_mu']=mu
+        ret['rep_theta']=theta
+        ret['rep_rho']=rho
+        ret['rep_alpha']=alpha
+        ret['rep_eta']=eta
+        ret['misc_q']=q
 #        ret['u0']=u0
 #        ret['uT']=uT
-        ret['v0']=v0
-        ret['vT']=vT
-        ret['vpath']=vpath
-        ret['upath']=self.upath
+        ret['rep_v0']=v0
+        ret['rep_vT']=vT
+        ret['ts_vpath']=vpath
+        ret['ts_upath']=self.upath
         
         return ret
 
@@ -370,8 +457,8 @@ class Heston_tree(svol_model):
 #---------------------------------------------
 
 class Heston_treeX2(svol_model):
-    def __init__( self, series,dt, model, method ):
-        super().__init__(series,dt, model, method)
+    def __init__( self, series,dt, model, method,options ):
+        super().__init__(series,dt, model, method,options)
 
         mu=0.0
         sigma=0.1
@@ -392,30 +479,117 @@ class Heston_treeX2(svol_model):
 #                 [hmu, hsigma, rho, alpha, xi,u0]
         self.workingpars_bounds=[(-0.5,0.5), (0.05, 0.5), (-0.9,0.9), (alpha_min, 20.0), (0.1, 4.5),(0.1,3.0)]
 
+        if 'init' in options:
+            self.initpars_reporting(options['init'])
+
 # precalculate anything that can absolutely be reused:
 #TODO: ugly!!
         Nret=self.Nobs-1
-        self.yasset=np.log( self.series[1:Nret+1]/self.series[0:Nret] )
-        self.upath=np.zeros(self.Nobs)
+        if(Nret>0):
+            self.yasset=np.log( self.series[1:Nret+1]/self.series[0:Nret] )
+            self.upath=np.zeros(self.Nobs)
 
 #TODO: expose these as options at some point?
-        self.ProbFactor =1.09
-        self.NormProbCalc='TDist'
-        
-        if( self.NormProbCalc == 'Normal' ):
-            sigma_base=np.std(self.yasset,ddof=1)
-            mu_base=np.mean(self.yasset)
-            self.cprob_base=norm.pdf( self.yasset, loc=mu_base, scale=sigma_base )
-        elif( self.NormProbCalc == 'TDist' ):
-            (t_df,t_loc,t_scale)=tdist.fit(self.yasset)
-            self.cprob_base=tdist.pdf(self.yasset,df=t_df,loc=t_loc,scale=t_scale)
-        else:
-            print('Unknown NormProbCalc -- Handle')
-    # allow for a multiplier:
-        self.cprob_base*=self.ProbFactor
+            self.ProbFactor =1.09
+            self.NormProbCalc='TDist'
+            
+            if( self.NormProbCalc == 'Normal' ):
+                sigma_base=np.std(self.yasset,ddof=1)
+                mu_base=np.mean(self.yasset)
+                self.cprob_base=norm.pdf( self.yasset, loc=mu_base, scale=sigma_base )
+            elif( self.NormProbCalc == 'TDist' ):
+                (t_df,t_loc,t_scale)=tdist.fit(self.yasset)
+                self.cprob_base=tdist.pdf(self.yasset,df=t_df,loc=t_loc,scale=t_scale)
+            else:
+                print('Unknown NormProbCalc -- Handle')
+        # allow for a multiplier:
+            self.cprob_base*=self.ProbFactor
 
         return
 
+    def initpars_reporting(self,pardict):
+
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+        u0=self.workingpars[5]
+
+        theta=sigma*sigma
+        eta=xi*sigma
+        v0=theta*u0
+        
+        for x in pardict:
+            if( x=='mu' ):
+                mu=pardict[x]
+            if( x=='theta' ):
+                theta=pardict[x]
+            if( x=='rho' ):
+                rho=pardict[x]
+            if( x=='alpha' ):
+                alpha=pardict[x]
+            if( x=='eta' ):
+                eta=pardict[x]
+            if( x=='v0' ):
+                v0=pardict[x]
+
+        sigma=np.sqrt(theta)
+        xi=eta/sigma
+        u0=v0/theta
+        
+        self.workingpars[0]=mu
+        self.workingpars[1]=sigma
+        self.workingpars[2]=rho
+        self.workingpars[3]=alpha
+        self.workingpars[4]=xi
+        self.workingpars[5]=u0
+
+        return
+
+    def get_structure(self):
+        assetname='asset'
+        variancename='variance'
+
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        u0=self.workingpars[5]
+
+        corrmatrix=np.array([[1.0,rho],[rho,1.0]])
+#TODO: best choice based on pars?
+        Nperstep=4
+
+        assetval=1.0
+        varianceval=u0*sigma*sigma
+
+        return assetname,assetval,variancename,varianceval,corrmatrix,Nperstep
+
+    def sim_step(self,asset,variance,Zs):
+
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+
+        theta=sigma*sigma
+        eta=xi*sigma
+        
+        Nperstep=np.shape(Zs)[1]
+        sim_asset=np.log(asset)
+        sim_variance =variance       
+        
+        dt=self.dt/Nperstep
+        vmin=1.0e-12
+        
+        for cc in range(0,Nperstep):
+            sim_asset+=(mu-0.5*sim_variance)*dt+np.sqrt(sim_variance*dt)*Zs[0,cc,:]
+            sim_variance=np.power(np.sqrt(sim_variance)+0.5*eta*np.sqrt(dt)*Zs[1,cc,:],2)+alpha*(theta-sim_variance)*dt-eta*eta*dt/4.0
+            sim_variance=np.maximum(sim_variance,vmin)
+
+        sim_asset=np.exp(sim_asset)
+        return sim_asset,sim_variance
+
     def get_reportingpars(self):
         super().get_reportingpars()
 
@@ -442,18 +616,18 @@ class Heston_treeX2(svol_model):
     
         vpath=sigma*sigma*self.upath
 
-        ret['mu']=mu
-        ret['theta']=theta
-        ret['rho']=rho
-        ret['alpha']=alpha
-        ret['eta']=eta
-        ret['q']=q
+        ret['rep_mu']=mu
+        ret['rep_theta']=theta
+        ret['rep_rho']=rho
+        ret['rep_alpha']=alpha
+        ret['rep_eta']=eta
+        ret['misc_q']=q
 #        ret['u0']=u0
 #        ret['uT']=uT
-        ret['v0']=v0
-        ret['vT']=vT
-        ret['vpath']=vpath
-        ret['upath']=self.upath
+        ret['rep_v0']=v0
+        ret['rep_vT']=vT
+        ret['ts_vpath']=vpath
+        ret['ts_upath']=self.upath
         
         return ret
 
@@ -792,8 +966,8 @@ class Heston_treeX2(svol_model):
 #---------------------------------------------
 
 class Heston_grid(svol_model):
-    def __init__( self, series,dt, model, method ):
-        super().__init__(series,dt, model, method)
+    def __init__( self, series,dt, model, method,options ):
+        super().__init__(series,dt, model, method,options)
 
         mu=0.0
         sigma=0.1
@@ -812,14 +986,98 @@ class Heston_grid(svol_model):
 #                 [hmu, hsigma, rho, alpha, xi,u0]
         self.workingpars_bounds=[(-0.5,0.5), (0.05, 0.5), (-0.9,0.9), (alpha_min, 20.0), (0.1, 4.0)]
 
+        if 'init' in options:
+            self.initpars_reporting(options['init'])
+
 # precalculate anything that can absolutely be reused:
 #TODO: ugly!!
         Nret=self.Nobs-1
-        self.yasset=np.log( self.series[1:Nret+1]/self.series[0:Nret] )
-        self.upath=np.zeros(self.Nobs)
+        if(Nret>0):
+            self.yasset=np.log( self.series[1:Nret+1]/self.series[0:Nret] )
+            self.upath=np.zeros(self.Nobs)
 
         return
 
+    def initpars_reporting(self,pardict):
+
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+
+        theta=sigma*sigma
+        eta=xi*sigma
+
+        for x in pardict:
+            if( x=='mu' ):
+                mu=pardict[x]
+            if( x=='theta' ):
+                theta=pardict[x]
+            if( x=='rho' ):
+                rho=pardict[x]
+            if( x=='alpha' ):
+                alpha=pardict[x]
+            if( x=='eta' ):
+                eta=pardict[x]
+# this sucks: the model needs a v0, but the optimiation doesn't...
+            if( x=='v0' ):
+                v0=pardict[x]
+
+        sigma=np.sqrt(theta)
+        xi=eta/sigma
+
+        self.workingpars[0]=mu
+        self.workingpars[1]=sigma
+        self.workingpars[2]=rho
+        self.workingpars[3]=alpha
+        self.workingpars[4]=xi
+        self.frozen_u0=v0/theta
+
+        return
+
+    def get_structure(self):
+        assetname='asset'
+        variancename='variance'
+
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+
+        corrmatrix=np.array([[1.0,rho],[rho,1.0]])
+#TODO: best choice based on pars?
+        Nperstep=4
+
+        assetval=1.0
+        varianceval=self.frozen_u0*sigma*sigma
+
+        return assetname,assetval,variancename,varianceval,corrmatrix,Nperstep
+
+    def sim_step(self,asset,variance,Zs):
+
+        mu=self.workingpars[0] 
+        sigma=self.workingpars[1]
+        rho=self.workingpars[2]
+        alpha=self.workingpars[3] 
+        xi=self.workingpars[4]
+
+        theta=sigma*sigma
+        eta=xi*sigma
+        
+        Nperstep=np.shape(Zs)[1]
+        sim_asset=np.log(asset)
+        sim_variance =variance       
+        
+        dt=self.dt/Nperstep
+        vmin=1.0e-12
+        
+        for cc in range(0,Nperstep):
+            sim_asset+=(mu-0.5*sim_variance)*dt+np.sqrt(sim_variance*dt)*Zs[0,cc,:]
+            sim_variance=np.power(np.sqrt(sim_variance)+0.5*eta*np.sqrt(dt)*Zs[1,cc,:],2)+alpha*(theta-sim_variance)*dt-eta*eta*dt/4.0
+            sim_variance=np.maximum(sim_variance,vmin)
+
+        sim_asset=np.exp(sim_asset)
+        return sim_asset,sim_variance
+
     def get_reportingpars(self):
         super().get_reportingpars()
 
@@ -833,6 +1091,7 @@ class Heston_grid(svol_model):
 
         theta=sigma*sigma
         eta=xi*sigma
+
         q=2.0*alpha/(xi*xi)
 
         self.variancepath()
@@ -845,18 +1104,18 @@ class Heston_grid(svol_model):
     
         vpath=sigma*sigma*self.upath
 
-        ret['mu']=mu
-        ret['theta']=theta
-        ret['rho']=rho
-        ret['alpha']=alpha
-        ret['eta']=eta
-        ret['q']=q
+        ret['rep_mu']=mu
+        ret['rep_theta']=theta
+        ret['rep_rho']=rho
+        ret['rep_alpha']=alpha
+        ret['rep_eta']=eta
+        ret['misc_q']=q
 #        ret['u0']=u0
 #        ret['uT']=uT
-        ret['v0']=v0
-        ret['vT']=vT
-        ret['vpath']=vpath
-        ret['upath']=self.upath
+        ret['rep_v0']=v0
+        ret['rep_vT']=vT
+        ret['ts_vpath']=vpath
+        ret['ts_upath']=self.upath
         
         return ret
     
